# Copyright © 2023 Neven Sajko

# Derivative-free root finder for a unidimensional function.

module BisectionMethod

struct E{F} <: Exception
  ab::NTuple{2, F}
end

function search(fun::Fun, ab::NTuple{2, F}) where {Fun<:Function, F<:AbstractFloat}
  (a, b) = minmax(ab...)
  sign_bits = map(signbit ∘ fun, (a, b))
  ((a != b) & !=(sign_bits...) & all(isfinite, (a, b))) || throw(E(ab))
  h = one(a)/2
  sb = first(sign_bits)

  while true
    m = fma(h, a, h*b)  # midpoint
    (a < m < b) || break

    if signbit(fun(m)) == sb
      a = m
    else
      b = m
    end
  end

  while true
    m = nextfloat(a)
    (m < b) || break

    if signbit(fun(m)) == sb
      a = m
    else
      b = m
    end
  end

  b
end

end
