# Copyright © 2023 Neven Sajko

module FindRoots

import
  ..CollectionWithReserved,
  ..LinearInterpolation,
  ..BisectionMethod

const coll_reserved = CollectionWithReserved.collection_with_reserved

function extract_brackets(c::AbstractVector{Tuple{F, Bool}}) where {F<:AbstractFloat}
  ret = NTuple{2, F}[]

  for i ∈ 0:(length(c) - 2)
    l = c[begin + i]
    r = c[begin + i + 1]
    (last(l) == last(r)) || push!(ret, (first(l), first(r)))
  end

  ret
end

function function_sign_bits(
  fun::Fun, interval::Itv, granularity::Int, additional_points::AbstractVector{F},
) where {Fun<:Function, F<:AbstractFloat, Itv<:NTuple{2, F}}
  gran = 1 << granularity

  lerpi = let m = gran, itv = interval
    i -> LinearInterpolation.lerpi(i, m, itv...)
  end

  point_with_signbit = let f = fun
    p -> (p, signbit(f(p)))
  end

  sign_bits = coll_reserved(Vector{Tuple{F, Bool}}, gran + 1 + length(additional_points))

  for i ∈ 0:gran
    push!(sign_bits, point_with_signbit(lerpi(i)))
  end

  for a ∈ additional_points
    push!(sign_bits, point_with_signbit(a))
  end

  unique!(sort!(sign_bits, by = first))
end

function_root_brackets(
  fun::Fun, interval::Itv, granularity::Int, additional_points::AbstractVector{F},
) where {Fun<:Function, F<:AbstractFloat, Itv<:NTuple{2, F}} =
  extract_brackets(function_sign_bits(fun, interval, granularity, additional_points))

function find_roots(
  fun::Fun, interval::Itv, granularity::Int,
  additional_bracketing_points::AbstractVector{F} = F[],
) where {Fun<:Function, F<:AbstractFloat, Itv<:NTuple{2, F}}
  find_root = let f = fun
    (ab::NTuple{2, F}) -> BisectionMethod.search(f, ab)
  end
  map(find_root, function_root_brackets(fun, interval, granularity, additional_bracketing_points))
end

end
