# Copyright © 2023 Neven Sajko. All rights reserved.

module IntervalHelpers

import ..NumericalErrorTypes

const NE = NumericalErrorTypes

interval_given_max_abs_err(val::AbstractFloat, max_err::Real) =
  (val - max_err, val + max_err)

interval_given_max_abs_err(max_err::Real) =
  let e = max_err
    v -> interval_given_max_abs_err(v, e)
  end

function intervals_given_max_abs_err(
  vals::AbstractVector{<:AbstractFloat},
  max_err::Real)

  (0 < max_err) || error("max error not positive")
  map(interval_given_max_abs_err(max_err), vals)
end

interval_given_max_rel_err(val::F, max_err::Real) where {F <: AbstractFloat} =
  interval_given_max_abs_err(
    val,
    max_err * abs(val),
  )

interval_given_max_rel_err(max_err::Real) =
  let e = max_err
    v -> interval_given_max_rel_err(v, e)
  end

function intervals_given_max_rel_err(
  vals::AbstractVector{<:AbstractFloat},
  max_err::Real)

  (0 < max_err) || error("max error not positive")
  map(interval_given_max_rel_err(max_err), vals)
end

intervaler_given_error_type(::NE.AbsoluteError) =
  intervals_given_max_abs_err

intervaler_given_error_type(::NE.RelativeError) =
  intervals_given_max_rel_err

intervals_given_max_err(
  vals::AbstractVector{<:AbstractFloat},
  error_type::NE.NumericalErrorType,
  max_err::Real) =
  intervaler_given_error_type(error_type)(vals, max_err)

end
