# Copyright 2022 Neven Sajko. All right reserved.

module LinCombs

export
  Handle, LinComb, dense, variables, to_dense

struct Handle
  h::Int32
end

Base.isless(l::Handle, r::Handle) = l.h < r.h

# Linear combination
const LinComb = Vector{Pair{Handle, Coef}} where {Coef <: Number}

dense(var_cnt::Int) =
  let c = var_cnt
    function(lc::LinComb{F}) where {F <: Real}
      local ret = zeros(F, c)
      for term in lc
        ret[first(term).h] = last(term)
      end
      ret
    end
  end

min_var(c::LinComb{<:Real}) = minimum(term -> first(term).h, c)
max_var(c::LinComb{<:Real}) = maximum(term -> first(term).h, c)

function dense(v::AbstractVector{<:LinComb{<:Real}})
  local min = minimum(min_var, v)
  local max = maximum(max_var, v)

  isone(min) || error("unexpected")

  local var_cnt = Int(max)
  map(dense(var_cnt), v)
end

variables(::Type{F}, count::Int) where {F <: Real} =
  LinComb{F}[[Handle(i) => one(F)] for i in 1:count]

function to_dense(::Type{T}, m::Vector{<:LinComb{<:Any}}) where {T <: Any}
  # Assuming m is square
  local n = length(m)

  local ret = zeros(T, n, n)

  for i in eachindex(m)
    local row = m[i]
    for e in row
      local j = first(e).h
      ret[i, j] = last(e)
    end
  end

  ret
end

end
