# Copyright © 2023 Neven Sajko

module LinearInterpolation

# Linear interpolation
function lerpi(i::Int, n::Int, a::F, b::F) where {F <: AbstractFloat}
  t = F(i) / n
  fma(t, b, fma(-t, a, a))
end

# Makes a grid between `a` and `b`, not including either endpoint, using linear
# interpolation. The number of elements in the returned vector is `n`.
function evenly_spaced_grid(n::Int, a::F, b::F) where {F <: AbstractFloat}
  m = n + 1
  F[lerpi(i, m, a, b) for i ∈ 1:(m - 1)]
end

end
