# Copyright © 2023 Neven Sajko

module NumericalErrors

import ..NumericalErrorTypes

const NE = NumericalErrorTypes

numerical_error_unchecked(::NE.AbsoluteErrorSigned, accurate, approximate) =
  accurate - approximate

numerical_error_unchecked(::NE.RelativeErrorSigned, accurate, approximate) =
  true - if accurate == approximate
    zero(approximate)/one(accurate)
  else
    approximate/accurate
  end

numerical_error_unchecked_abs(et::NE.NumericalErrorType, accurate, approximate) =
  abs(numerical_error_unchecked(signed(et), accurate, approximate))

numerical_error_unchecked(et::NE.AbsoluteError, a, b) = numerical_error_unchecked_abs(et, a, b)
numerical_error_unchecked(et::NE.RelativeError, a, b) = numerical_error_unchecked_abs(et, a, b)

numerical_error_unchecked(::NE.RelativeDistance, a, b) =
  if a == b
    zero(a)
  elseif !iszero(a) && !iszero(b) && (signbit(a) == signbit(b))
    abs(log(a/b))
  else
    oftype(a, Inf)
  end

function numerical_error(error_type::NE.NumericalErrorType, accurate, approximate)
  isfinite(accurate) || error("accurate not finite")
  isfinite(approximate) || error("approximate not finite")
  numerical_error_unchecked(error_type, accurate, approximate)
end

end
