# Copyright © 2023 Neven Sajko

module SortedBoundedLengthVectors

import ..CollectionWithReserved

const coll_reserved = CollectionWithReserved.collection_with_reserved

const Order = Base.Order

struct SortedBoundedLengthVector{T <: Any, O <: Order.Ordering}
  max_len::Int

  sorted::Vector{T}

  order::O

  function SortedBoundedLengthVector{T, O}(
    max_len::Int, order::O,
  ) where {T <: Any, O <: Order.Ordering}
    v = coll_reserved(Vector{T}, max_len)
    new{T, O}(max_len, v, order)
  end
end

SortedBoundedLengthVector{T}(max_len::Int, order::Order.Ordering) where {T <: Any} =
  SortedBoundedLengthVector{T, typeof(order)}(max_len, order)

Base.isempty(v::SortedBoundedLengthVector) = isempty(v.sorted)

function linear_search(v::SortedBoundedLengthVector{T}, e::T) where {T}
  index = firstindex(v.sorted)
  while (index ≤ lastindex(v.sorted)) && !Order.lt(v.order, e, v.sorted[index])
    index += true
  end
  index
end

Base.length(v::SortedBoundedLengthVector) = Base.length(v.sorted)

Base.last(v::SortedBoundedLengthVector) = Base.last(v.sorted)

Base.pop!(v::SortedBoundedLengthVector) = Base.pop!(v.sorted)

is_full(v::SortedBoundedLengthVector) = v.max_len == length(v)

function Base.push!(v::SortedBoundedLengthVector{T}, e::T) where {T}
  index = linear_search(v, e)
  after = index == lastindex(v.sorted) + true
  full = is_full(v)
  (!after & full) && pop!(v)
  (!after | !full) && insert!(v.sorted, index, e)
  v
end

end
