# Copyright © 2023 Neven Sajko

module TinyFilter

import
  ..LinearInterpolation,
  ..FindRoots

const find_roots = FindRoots.find_roots

function itv_negative_ratio(
  fun::Fun, interval::Itv, granularity::Int,
) where {Fun<:Function, F<:AbstractFloat, Itv<:NTuple{2, F}}
  gran = 1 << granularity

  lerpi = let m = gran, itv = interval
    i -> LinearInterpolation.lerpi(i, m, itv...)
  end

  count = 0

  for i ∈ 1:(gran - 1)
    count += signbit(fun(lerpi(i)))
  end

  count / (gran - 1)
end

interval_is_tiny(
  fun::Fun, interval::Itv, granularity::Int,
) where {Fun<:Function, F<:AbstractFloat, Itv<:NTuple{2, F}} =
  1e-2 < itv_negative_ratio(fun, interval, granularity)

function tiny_filter!(
  fun::Fun, output::V, interval::Itv, granularity::Int,
  extra_bracketing_samples::AbstractVector{F},
) where {Fun<:Function, F<:AbstractFloat, Itv<:NTuple{2, F}, V<:AbstractVector{Itv}}
  roots = find_roots(fun, interval, granularity, extra_bracketing_samples)

  itv_is_tiny = let f = fun
    itv -> interval_is_tiny(f, itv, 9)
  end

  if isempty(roots)
    itv_is_tiny(interval) || push!(output, interval)
  else
    let it = (first(interval), first(roots))
      itv_is_tiny(it) || push!(output, it)
    end

    for i ∈ 0:(length(roots) - 2)
      it = (roots[begin + i], roots[begin + i + 1])
      itv_is_tiny(it) || push!(output, it)
    end

    let it = (last(roots), last(interval))
      itv_is_tiny(it) || push!(output, it)
    end
  end

  nothing
end

function tiny_filter(
  fun::Fun, itvs::AbstractVector{NTuple{2, F}}, max_tiny::Real, granularity::Int,
) where {Fun<:Function, F<:AbstractFloat}
  ret = NTuple{2, F}[]

  a = abs ∘ fun

  g = let a = a, mt = max_tiny
    x -> (a(x) - mt)::F
  end

  for itv ∈ itvs
    # TODO: also use the minima of `a` as extra bracketing sample points.
    #       Finding the points will be on a low effort basis.
    extra_bracketing_samples = find_roots(fun, itv, granularity)

    tiny_filter!(g, ret, itv, granularity, extra_bracketing_samples)
  end

  ret
end

end
