# Copyright © 2023 Neven Sajko. All rights reserved.

# TODO: @inferred, JET

import FindMinimaxPolynomial

using Test

const FMP = FindMinimaxPolynomial
const ItvH = FMP.IntervalHelpers

function test_infinite_error_interval(
  err::Fun,
  ::Type{F} = Float64,
) where {Fun <: Function, F <: AbstractFloat}
  @test (typemin(F), typemax(F)) == err(rand(F), typemax(F))
  nothing
end

function test_usual_magnitude_error_interval_abs_err(
  ::Type{F} = Float64,
) where {F <: AbstractFloat}
  err = rand(F) / 64
  val = (rand(F) + one(F)) / 4
  (a, b) = ItvH.interval_given_max_rel_err(val, err)
  @test zero(F) ≤ a < b ≤ one(F)
  nothing
end

function test_usual_magnitude_error_interval_rel_err(
  ::Type{F} = Float64,
) where {F <: AbstractFloat}
  (a, b) = ItvH.interval_given_max_rel_err(rand(F), rand(F))
  @test zero(F) ≤ a < b ≤ one(F)
  nothing
end

@testset "FindMinimaxPolynomial.jl" begin
  @test isempty(detect_ambiguities(FMP, recursive = true))

  @testset "IntervalHelpers" begin
    @testset "infinite error" begin
      @testset "absolute" for i ∈ 1:200
        test_infinite_error_interval(ItvH.interval_given_max_abs_err)
      end

      @testset "relative" for i ∈ 1:200
        test_infinite_error_interval(ItvH.interval_given_max_rel_err)
      end
    end

    @testset "usual magnitude error" begin
      @testset "absolute" for i ∈ 1:200
        test_usual_magnitude_error_interval_abs_err()
      end

      @testset "relative" for i ∈ 1:200
        test_usual_magnitude_error_interval_rel_err()
      end
    end
  end
end
